/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 2011-2017 OpenFOAM Foundation
     \\/     M anipulation  | Copyright (C) 2015-2016 OpenCFD Ltd.
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM.  If not, see <http://www.gnu.org/licenses/>.

\*---------------------------------------------------------------------------*/

#include "sampledWaveElevation.H"
#include "dictionary.H"
#include "Time.H"
#include "volFields.H"
#include "ListListOps.H"
#include "SortableList.H"
#include "volPointInterpolation.H"
#include "mapPolyMesh.H"
#include "addToRunTimeSelectionTable.H"

// * * * * * * * * * * * * * * Static Data Members * * * * * * * * * * * * * //

namespace Foam
{
namespace functionObjects
{
    defineTypeNameAndDebug(sampledWaveElevation, 0);

    addToRunTimeSelectionTable
    (
        functionObject,
        sampledWaveElevation,
        dictionary
    );
}
}


// * * * * * * * * * * * * * Private Member Functions  * * * * * * * * * * * //

void Foam::functionObjects::sampledWaveElevation::combineSampledSets
(
    PtrList<coordSet>& masterSampledSets,
    labelListList& indexSets
)
{
    // Combine sampleSets from processors. Sort by curveDist. Return
    // ordering in indexSets.
    // Note: only master results are valid

    masterSampledSets_.clear();
    masterSampledSets_.setSize(size());
    indexSets_.setSize(size());

    const PtrList<sampledSet>& surfaceElevation = *this;

    forAll(surfaceElevation, setI)
    {
        const sampledSet& samplePts = surfaceElevation[setI];

        // Collect data from all processors
        List<List<point>> gatheredPts(Pstream::nProcs());
        gatheredPts[Pstream::myProcNo()] = samplePts;
        Pstream::gatherList(gatheredPts);

        List<labelList> gatheredSegments(Pstream::nProcs());
        gatheredSegments[Pstream::myProcNo()] = samplePts.segments();
        Pstream::gatherList(gatheredSegments);

        List<scalarList> gatheredDist(Pstream::nProcs());
        gatheredDist[Pstream::myProcNo()] = samplePts.curveDist();
        Pstream::gatherList(gatheredDist);

        // Combine processor lists into one big list.
        List<point> allPts
        (
            ListListOps::combine<List<point>>
            (
                gatheredPts, accessOp<List<point>>()
            )
        );
        labelList allSegments
        (
            ListListOps::combine<labelList>
            (
                gatheredSegments, accessOp<labelList>()
            )
        );
        scalarList allCurveDist
        (
            ListListOps::combine<scalarList>
            (
                gatheredDist, accessOp<scalarList>()
            )
        );


        if (Pstream::master() && allCurveDist.size() == 0)
        {
            WarningInFunction
                << "Sample set " << samplePts.name()
                << " has zero points." << endl;
        }

        // Sort curveDist and use to fill masterSamplePts
        SortableList<scalar> sortedDist(allCurveDist);
        indexSets[setI] = sortedDist.indices();

        masterSampledSets.set
        (
            setI,
            new coordSet
            (
                samplePts.name(),
                samplePts.axis(),
                List<point>(UIndirectList<point>(allPts, indexSets[setI])),
                sortedDist
            )
        );
    }
}

void Foam::functionObjects::sampledWaveElevation::combineSampledValues
(
    const PtrList<volFieldSampler>& sampledFields,
    const labelListList& indexSets,
    PtrList<volFieldSampler>& masterFields
)
{
    forAll(sampledFields, fieldi)
    {
        List<Field<scalar>> masterValues(indexSets.size());

        forAll(indexSets, setI)
        {
            // Collect data from all processors
            List<Field<scalar>> gatheredData(Pstream::nProcs());
            gatheredData[Pstream::myProcNo()] = sampledFields[fieldi][setI];
            Pstream::gatherList(gatheredData);

            if (Pstream::master())
            {
                Field<scalar> allData
                (
                    ListListOps::combine<Field<scalar>>
                    (
                        gatheredData,
                        Foam::accessOp<Field<scalar>>()
                    )
                );

                masterValues[setI] = UIndirectList<scalar>
                (
                    allData,
                    indexSets[setI]
                )();
            }
        }

        masterFields.set
        (
            fieldi,
            new volFieldSampler
            (
                masterValues,
                sampledFields[fieldi].name()
            )
        );
    }
}

// * * * * * * * * * * * * * * * * Constructors  * * * * * * * * * * * * * * //

Foam::functionObjects::sampledWaveElevation::sampledWaveElevation
(
    const word& name,
    const Time& runTime,
    const dictionary& dict
)
:
    functionObject(name),
    logFiles(runTime, name),
    PtrList<sampledSet>(),
    mesh_
    (
        refCast<const fvMesh>
        (
            runTime.lookupObject<objectRegistry>
            (
                dict.lookupOrDefault("region", polyMesh::defaultRegion)
            )
        )
    ),
    dict_(dict),
    searchEngine_(mesh_),
    interpolationScheme_(word::null)
{
    read(dict_);
    resetName(typeName);
}

// * * * * * * * * * * * * * * * * Destructor  * * * * * * * * * * * * * * * //

// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

bool Foam::functionObjects::sampledWaveElevation::execute()
{
    return true;
}


bool Foam::functionObjects::sampledWaveElevation::write()
{
    logFiles::write();

    bool interpolate = interpolationScheme_ != "cell";

    // Storage for interpolated values - brute force size to one!
    PtrList<volFieldSampler> sampledFields(1);

    if (interpolate)
    {
        sampledFields.set
        (
            0,
            new volFieldSampler
            (
                interpolationScheme_,
                mesh_.lookupObject
                <GeometricField<scalar, fvPatchField, volMesh>>
                (alphaFieldName_),
                *this
            )
        );
    }
    else
    {
        sampledFields.set
        (
            0,
            new volFieldSampler
            (
                mesh_.lookupObject
                <GeometricField<scalar, fvPatchField, volMesh>>
                (alphaFieldName_),
                *this
            )
        );
    }

    // Combine sampled fields from processors.
    // Note: only master results are valid
    PtrList<volFieldSampler> masterFields(sampledFields.size());
    combineSampledValues(sampledFields, indexSets_, masterFields);

    if (Pstream::master())
    {
        scalarField result(masterSampledSets_.size(), 0.0);

        forAll (masterSampledSets_, seti)
        {
            const coordSet& samplePositions(masterSampledSets_[seti]);

            const Field<scalar>& sampleValues(masterFields[0][seti]);

            scalar tolerance(0.0001);

            if (sampleValues.size() < 2)
            {
                result[seti] = -GREAT;
            }
            else if
            (
                (
                    sampleValues[0] < tolerance &&
                    sampleValues[sampleValues.size()-1] < tolerance
                )
                ||
                (
                    sampleValues[0] > 1.0 - tolerance &&
                    sampleValues[sampleValues.size()-1] > 1.0 - tolerance
                )
            )
            {
                result[seti] = -GREAT;
            }
            else
            {
                // lowest sample position
                scalar eta0(samplePositions.scalarCoord(0));

                // wave elevation above the eta0
                scalar eta(0);

                // integration over z direction, \int_{eta0} \alpha dz,
                // makes wave elevation above eta0
                // since \alpha is a fraction of volume of fluid
                for (int pointi=0; pointi < sampleValues.size() - 1; pointi++)
                {
                    eta +=
                        (
                            samplePositions.scalarCoord(pointi + 1)
                            - samplePositions.scalarCoord(pointi)
                        )
                        *(sampleValues[pointi + 1] + sampleValues[pointi])*0.5;

                    eta0 =
                        Foam::min
                        (
                            eta0,
                            samplePositions.scalarCoord(pointi + 1)
                        );
                }

                result[seti] = eta + eta0;
            }
        }

        writeTime(file());

        // Write the surface elevation
        forAll (result, seti)
        {
            file() << tab << result[seti];
        }

        file() << endl;
    }

    return true;
}


void Foam::functionObjects::sampledWaveElevation::writeFileHeader(const label)
{
    OFstream& file = this->file();

    writeHeader(file, "SurfaceElevation");
    file<< setw(charWidth()) << "Time" << tab;
    forAll(masterSampledSets_, seti)
    {
        file<< masterSampledSets_[seti].name() << tab;
    }
    file << endl;
}

bool Foam::functionObjects::sampledWaveElevation::read(const dictionary& dict)
{
    bool setsFound = dict_.found("sets");

    if (setsFound)
    {
        dict_.lookup("interpolationScheme") >> interpolationScheme_;
        alphaFieldName_ = dict_.lookupOrDefault<word>("alphaFieldName", "alpha.water");

        PtrList<sampledSet> newList
        (
            dict_.lookup("sets"),
            sampledSet::iNew(mesh_, searchEngine_)
        );

        transfer(newList);
        combineSampledSets(masterSampledSets_, indexSets_);

        if (this->size())
        {
            Info<< "Reading set description:" << nl;
            forAll(*this, setI)
            {
                Info<< "    " << operator[](setI).name() << nl;
            }
            Info<< endl;
        }
    }

    return true;
}


void Foam::functionObjects::sampledWaveElevation::correct()
{
    bool setsFound = dict_.found("sets");

    if (setsFound)
    {
        searchEngine_.correct();

        PtrList<sampledSet> newList
        (
            dict_.lookup("sets"),
            sampledSet::iNew(mesh_, searchEngine_)
        );
        transfer(newList);
        combineSampledSets(masterSampledSets_, indexSets_);
    }
}


void Foam::functionObjects::sampledWaveElevation::updateMesh(const mapPolyMesh& mpm)
{
    if (&mpm.mesh() == &mesh_)
    {
        correct();
    }
}


void Foam::functionObjects::sampledWaveElevation::movePoints(const polyMesh& mesh)
{
    if (&mesh == &mesh_)
    {
        correct();
    }
}


void Foam::functionObjects::sampledWaveElevation::readUpdate(const polyMesh::readUpdateState state)
{
    if (state != polyMesh::UNCHANGED)
    {
        correct();
    }
}



// ************************************************************************* //
