/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 2011-2016 OpenFOAM Foundation
     \\/     M anipulation  | Copyright (C) 2015-2017 OpenCFD Ltd.
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM.  If not, see <http://www.gnu.org/licenses/>.

Class
    Foam::functionObjects::waveElevation

Description
    Writes the wave elevation.
    wave elevation is sampled by sampledSet and integrated along z direction.
    Default field for sampling is alpha.water from VoF field.

Usage
    waveElevation
    {
        type               waveElevation;

        writeControl      timeStep;
        writeInterval     1;

        interpolationScheme  cellPointFace;
        alphaFieldName       alpha.water;

        sets
        (
             gauge0
             {
                 type lineUniform;
                 axis z;
                 start (2  2  -10);
                 end   (2  2  10);
                 nPoints 100;
             }
        );
    }

SourceFiles
    waveElevation.C
    volFieldSampler.C

\*---------------------------------------------------------------------------*/

#ifndef sampledWaveElevation_H
#define sampledWaveElevation_H

#include "functionObject.H"
#include "logFiles.H"
#include "sampledSet.H"
#include "volFieldsFwd.H"
#include "meshSearch.H"
#include "interpolation.H"
#include "coordSet.H"
#include "writer.H"
#include "wordReList.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{

// Forward declaration of classes
class Time;
class objectRegistry;
class dictionary;
class fvMesh;

namespace functionObjects
{
/*---------------------------------------------------------------------------*\
                         Class surfaceElevation Declaration
\*---------------------------------------------------------------------------*/

class sampledWaveElevation
:
    public functionObject,
    public logFiles,
    public PtrList<sampledSet>
{
        //- Class used for sampling volFields

        class volFieldSampler
        :
            public List<Field<scalar>>
        {
            //- Name of this collection of values
            const word name_;

        public:

            //- Construct interpolating field to the sampleSets
            volFieldSampler
            (
                const word& interpolationScheme,
                const GeometricField<scalar, fvPatchField, volMesh>& field,
                const PtrList<sampledSet>&
            );

            //- Construct mapping field to the sampleSets
            volFieldSampler
            (
                const GeometricField<scalar, fvPatchField, volMesh>& field,
                const PtrList<sampledSet>&
            );

            //- Construct from components
            volFieldSampler
            (
                const List<Field<scalar>>& values,
                const word& name
            );

            //- Return the field name
            const word& name() const { return name_; }
        };

        // Private data

        scalarField alphaField_;

        word alphaFieldName_;

        PtrList<sampledSet> sets;

        //- Const reference to fvMesh
        const fvMesh &mesh_;

        //- Keep the dictionary to recreate sets for moving mesh cases
        dictionary dict_;

        //- Mesh search engine
        meshSearch searchEngine_;

        word interpolationScheme_;

        // Merging structures
        PtrList<coordSet> masterSampledSets_;
        labelListList indexSets_;

        // Private Member Functions

        //- Combine points from all processors. Sort by curveDist and produce
        //  index list. Valid result only on master processor.
        void combineSampledSets
        (
            PtrList<coordSet> &masterSampledSets,
            labelListList &indexSets
        );

        //- Combine values from all processors.
        //  Valid result only on master processor.
        void combineSampledValues
        (
            const PtrList<volFieldSampler> &sampledFields,
            const labelListList &indexSets,
            PtrList<volFieldSampler> &masterFields
        );

        //- Disallow default bitwise copy construct and assignment
        sampledWaveElevation(const sampledWaveElevation &) = delete;
        void operator=(const sampledWaveElevation &) = delete;

      public:
        //- Runtime type information
        TypeName("waveElevation");

        // Constructors
        //- Construct from Time and dictionary
        sampledWaveElevation
        (
            const word& name,
            const Time& time,
            const dictionary& dict
        );

        //- Destructor
        virtual ~sampledWaveElevation() = default;

        // Member Functions

        virtual void writeFileHeader(const label i = 0);

        //- Read the surfaceElevation
        virtual bool read(const dictionary& dict);

        //- Execute, currently does nothing
        virtual bool execute();

        //- Sample and write
        virtual bool write();

        //- Correct for mesh changes
        void correct();

        //- Update for changes of mesh
        virtual void updateMesh(const mapPolyMesh& mpm);

        //- Update for mesh point-motion
        virtual void movePoints(const polyMesh& mesh);

        //- Update for changes of mesh due to readUpdate
        virtual void readUpdate(const polyMesh::readUpdateState state);
};

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace functionObjects
} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#endif

// ************************************************************************* //
